/* $Id: keyplex.c,v 1.10 2007/03/08 20:15:46 drwilco Exp $ */
/*

"THE BEER-WARE LICENSE" (Revision 43):
<drwilco@drwilco.net> wrote this file. As long as you retain this notice
you can do whatever you want with this stuff. If we meet some day, and you
think this stuff is worth it, you can buy me a drink in return.
                                                     Rogier R. Mulhuijzen

(I don't like beer, hence the raising of the revision. Thank you
Poul-Henning Kamp for this otherwise awesome license, and many other
contributions to the world of code.)

*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <ctype.h>
#include <X11/Xlocale.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>

#define WINDOW_WIDTH 200
#define WINDOW_HEIGHT 100
#define WINDOW_X 100
#define WINDOW_Y 100

/* #define DEBUG */
#ifdef DEBUG
#define DEBUGPRINT(...) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUGPRINT(...) 
#endif

extern char **environ;
                                
const char *program;
Display *dpy;
int screen;
int num = 2;
int numindex = 0;
Window *windows;
Window *pasteWindows;
char *keyplexid;
char *termcmd = "uxterm -xrm 'xterm*allowSendEvents: true' -class %s";
Bool querying = False;
Bool sendingEvents = False;
XErrorHandler oldHandler;

int errorHandler(Display *dpy, XErrorEvent *error) {
  if (error->error_code = BadWindow) {
    DEBUGPRINT("BadWindow\n");
    if (querying) {
      /* Window disappeared while querying the whole tree.
       * Not a problem really */
      DEBUGPRINT("During XQueryTree, no prob\n");
      return 0;
    }
    if (sendingEvents) {
      /* Window gone, let's take it out of the list of windows
             * we're sending events to. */
      Window *newWindows;
      Window *newPasteWindows;
      int i, j;

      DEBUGPRINT("During Event Sending, gotta remove\n");
      if (num == 1) {
        fprintf(stderr, "%s: My last window disappeared, exiting.\n",
                program);
                 exit(0);
      }
            
      newWindows = malloc(sizeof(Window) * (num-1));
      newPasteWindows = malloc(sizeof(Window) * (num-1));

      j = 0;
      for (i = 0; i < num; ++i) {
        if (j == num-1) {
          /* Window disappeared but we don't track it...
           * Maybe it's a sync thing. */
          DEBUGPRINT("didn't find the window to remove, that's weird\n");
          free(newWindows);
          free(newPasteWindows);
          return 0;
        }
        if (windows[i] != error->resourceid
            && pasteWindows[i] != error->resourceid) {
          DEBUGPRINT("Copying Window #%d\n", i);
          newWindows[j] = windows[i];
          newPasteWindows[j++] = pasteWindows[i];
        }
      }
      free(windows);
      free(pasteWindows);
      windows = newWindows;
      pasteWindows = newPasteWindows;
      --num;
      return 0;
    }
  }
  return oldHandler(dpy, error);
}

static void spawnSlaves() {
  int i, len;
  char *classbase = "keyplex";
  char *cmd;

  len=(sizeof(pid_t) * 2) + strlen(classbase);
  keyplexid=malloc(len);
  if (!keyplexid) {
    fprintf(stderr, "%s: malloc() failed\n", program , i);
    exit(6);
  }
  snprintf(keyplexid, len, "%s%x", classbase, getpid());
  DEBUGPRINT("Temp: %s\n", keyplexid);
  len=strlen(termcmd) + (sizeof(pid_t) * 2) + strlen(classbase);
  cmd=malloc(len);
  if (!cmd) {
    fprintf(stderr, "%s: malloc() failed\n", program , i);
    exit(6);
  }
  snprintf(cmd, len, termcmd, keyplexid);
  for (i = 0; i < num; ++i) {

    DEBUGPRINT("Spawning slave #%d: %s\n", i, cmd);
    pid_t pid = fork();
    if (pid == 0) {
      execl("/bin/sh", "sh", "-c", cmd, (char *)0);
      fprintf(stderr, "%s: Spawning slave #%d failed, errno: %d\n"
                    , program, i, errno);
      _exit(127);
    }
    if (pid == -1) {
      fprintf(stderr, "%s: Forking #%d failed\n", program , i);
    }
  }
  free(cmd);
}

void addSlave(Window w) {
  Window root, parent;
  int nchildren, i;
  long area = 0L;
  Window *children;
  XWindowAttributes wattr;

  if (numindex >= num) {
    fprintf(stderr, "%s: Found too many children\n", program);
    exit(2);
  }

  if (XQueryTree(dpy, w, &root, &parent, &children, &nchildren)) {
    /* DEBUGPRINT("Found %u children\n", nchildren); */
    for (i=0; i < nchildren; ++i) {
      if (XGetWindowAttributes(dpy, children[i], &wattr) 
        && wattr.width * wattr.height > area) {
        area = wattr.width * wattr.height;
        pasteWindows[numindex] = children[i];
      }
    }
    if (nchildren > 0) {
      XFree(children);
    }
    windows[numindex] = w;
    ++numindex;
  }
}

void findSlaves(Window w) {
  Window root, parent;
  int nchildren, i;
  Window *children;
  XWindowAttributes wattr;
  XClassHint classhint;


  if (XQueryTree(dpy, w, &root, &parent, &children, &nchildren)) {
    /* DEBUGPRINT("Found %u children\n", nchildren); */
    for (i=0; i < nchildren; ++i) {
      Bool added = False;
      if (XGetClassHint(dpy, children[i], &classhint)) {
        DEBUGPRINT("0x%x, %s, %s\n", children[i], classhint.res_class,
                   classhint.res_name);
        if (strcmp(classhint.res_class, keyplexid) == 0 
            || strcmp(classhint.res_name, keyplexid) == 0) {
          DEBUGPRINT("match on class/name\n");
          if (XGetWindowAttributes(dpy, children[i], &wattr)) {
            DEBUGPRINT("got wattr, checking Viewable (map_state=%d)\n",
                       wattr.map_state);
            if (wattr.map_state == IsViewable) {
              DEBUGPRINT("adding 0x%x\n", children[i]);
              addSlave(children[i]);
              DEBUGPRINT("found=%d/%d\n", numindex, num);
            }
          }
        }
      }
      if (!added) {
        findSlaves(children[i]);
      }
    }
    if (nchildren > 0) {
      XFree(children);
    } 
  }
}


static void sendKeyPress(XEvent *eventp) {
  XKeyEvent *e = (XKeyEvent *) eventp;
  int i;

  for (i = 0; i < num; ++i) {
    e->window=windows[i];
    DEBUGPRINT("Sending event to 0x%x @ %d\n", e->window, i);
    XSendEvent(e->display, e->window, False, KeyPressMask, (XEvent *)e);
  }

}

static void sendKeyRelease(XEvent *eventp) {
  XKeyEvent *e = (XKeyEvent *) eventp;
  int i;

  for (i = 0; i < num; ++i) {
    e->window=windows[i];
    DEBUGPRINT("Sending event to 0x%x @ %d\n", e->window, i);
    XSendEvent(e->display, e->window, False, KeyReleaseMask, (XEvent *)e);
  }

}


static void sendButtonPress(XEvent *eventp) {
  XButtonEvent *e = (XButtonEvent *) eventp;
  int i;

  if (e->button == 2) {
    for (i = 0; i < num; ++i) {
      e->window=pasteWindows[i];
      DEBUGPRINT("Sending event to 0x%x @ %d\n", e->window, i);
      XSendEvent(e->display, e->window, False, ButtonPressMask, (XEvent *)e);
    }
  }
}
            
static void sendButtonRelease(XEvent *eventp) {
  XButtonEvent *e = (XButtonEvent *) eventp;
  int i;

  if (e->button == 2) {
    for (i = 0; i < num; ++i) {
      e->window=pasteWindows[i];
      DEBUGPRINT("Sending event to 0x%x @ %d\n", e->window, i);
      XSendEvent(e->display, e->window, False, ButtonReleaseMask, (XEvent *)e);
    }
  }
}


static void usage()
{
  fprintf (stderr, "usage:  %s [<option> ...]\n", program);
  fprintf (stderr, "where options include:\n");
  fprintf (stderr, "  --display=<display>             X display to connect to\n");
  fprintf (stderr, "  --number=<number>               Number of terminals to spawn and control\n");
  fprintf (stderr, "  --command=<cmdline>             Command line to use to spawn a terminal\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "The command line must contain a %%s which will be replaced with a unique\n");
  fprintf (stderr, "string used to find the terminals in X. The %%s should be used as an argument\n");
  fprintf (stderr, "to set either the window class or name\n");
  fprintf (stderr, "Examples:\n");
  fprintf (stderr, "  %s --command \"xterm -xrm 'xterm*allowSendEvents: true' -class %%s\"\n", program);
  fprintf (stderr, "  %s --command \"gnome-terminal --disable-factory --class=%%s\"\n", program);
  fprintf (stderr, "  %s --command \"uxterm -xrm 'xterm*allowSendEvents: true' \" \\\n", program);
  fprintf (stderr, "                    \" -bg black -fg green -class %%s\"\n");
  fprintf (stderr, "  %s --command \"aterm -name %%s\"\n", program);
  fprintf (stderr, "\n");
  fprintf (stderr, "Note: xterm (and uxterm) require the resource allowSendEvents to be True.\n");
  fprintf (stderr, "      Add key \"-xrm 'xterm*allowSendEvents: true'\" to the command line\n");
  fprintf (stderr, "      or \"xterm*allowSendEvents: True\" to ~/.Xresources to set this permanently.\n");
  fprintf (stderr, "\n");
  exit(1);
}

int
main(int argc, char **argv)
{
  char *displayname = NULL;
  int i;
  Window w;
  XSetWindowAttributes attr;
  XWindowAttributes wattr;
  unsigned long mask = 0L;
  char *name = "Keyplex input";
  static struct option long_options[] = {
    {"display",   required_argument, 0, 'd'},
    {"id",        required_argument, 0, 'i'},
    {"number",    required_argument, 0, 'n'},
    {"command",   required_argument, 0, 'c'},
    {0, 0, 0, 0}
  };
  int option_index = 0;
  int option_val;

  program = argv[0];

  w = 0;
  while ((option_val = getopt_long(argc, argv, "d:g:i:n:c:",
                            long_options, &option_index)) != -1) {
    switch (option_val) {
      case 'd':
        displayname = optarg;
        continue;
      case 'i':
        sscanf(optarg, "0x%lx", &w);
        if (!w) {
          sscanf(optarg, "%lu", &w);
        }
        if (!w) {
          usage();
        }
        continue;
      case 'n':
        sscanf(optarg, "%lu", &num);
        continue;
      case 'c':
        termcmd = optarg;
        continue;
      default:
        usage();
    }
  }                                        /* end for over argc */

  spawnSlaves();

  dpy = XOpenDisplay (displayname);
  if (!dpy) {
    fprintf (stderr, "%s: unable to open display '%s'\n",
             program, XDisplayName (displayname));
    exit(1);
  }

  oldHandler = XSetErrorHandler(errorHandler);

  screen = DefaultScreen (dpy);

  windows = malloc(sizeof(Window) * num);
  pasteWindows = malloc(sizeof(Window) * num);
  querying = True;
  while (numindex < num) {
    usleep(250000);
    numindex = 0;
    findSlaves(RootWindow(dpy, screen));
  }
  querying = False;

  attr.event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask 
                      | ButtonReleaseMask;

  if (w) {
    XGetWindowAttributes(dpy, w, &wattr);
    if (wattr.all_event_masks & ButtonPressMask) {
      attr.event_mask &= ~ButtonPressMask;
    }
    attr.event_mask &= ~SubstructureRedirectMask;
    XSelectInput(dpy, w, attr.event_mask);
  } else {
    attr.background_pixel = WhitePixel(dpy,screen);
    attr.border_pixel = BlackPixel(dpy,screen);
    mask |= (CWBackPixel | CWBorderPixel | CWEventMask);

    w = XCreateWindow (dpy, RootWindow (dpy, screen), WINDOW_X, WINDOW_Y,
                       WINDOW_WIDTH, WINDOW_HEIGHT, 2, 0, InputOutput, 
                       (Visual *)CopyFromParent, mask, &attr);

    XSetStandardProperties (dpy, w, name, NULL, (Pixmap) 0, argv, argc,
                                NULL);

    XMapWindow (dpy, w);
  }

  sendingEvents = True;

  for (;;) {
    XEvent event;

    XNextEvent (dpy, &event);

    switch (event.type) {
      case KeyPress:
        sendKeyPress (&event);
        break;
      case KeyRelease:
        sendKeyRelease (&event);
        break;
      case ButtonPress:
        sendButtonPress (&event);
        break;
      case ButtonRelease:
        sendButtonRelease (&event);
        break;
    }
  }

  XCloseDisplay (dpy);
}
